var m_net = require('net');
var m_events = require('events');
m_fs = require('fs');
var m_cache;
var m_rest;
var m_utils;
var m_epad;
var m_dbus;

var moduleName = 'app_chromium';
var chromiumUrl;

var chromiumPort = '1233';
var reconnectTimeout = 6000;
var client = new m_net.Socket();
var chromiumData = null;
var cmdResult = null;
var autoRebootInterval = null;
var startTime = Date.now();
var responseEvent = new m_events.EventEmitter();

const status = {
    PENDING: -1,
    OK: 0,
    CMD_ERROR: 1,
    FORMAT_ERROR: 2
}
var cmdStatus = status.PENDING;

function autoRebootHandle() {

    // Clock sanity check. Year should be >= 2023
    if (new Date().getYear() < 123)
        return;

    // Skip if chromium is started less then one hour ago
    elapsed = Date.now() - startTime;
    if (elapsed > 0 && elapsed < 60*60*1000 )
        return;

    // Chromium not running
    if (client.destroyed)
        return;

    currMins = new Date().getHours()*60 + new Date().getMinutes()
    restartMins = chromiumData.settings.autoRestartTimeMins ? chromiumData.settings.autoRestartTimeMins : 0;

    // Push the time a bit ahead if the hour is near midnight to avoid issues with the clock wrapping around.
    restartMins = restartMins > 60*24 - 3 ? 0 : restartMins;

    deltaMins = currMins - restartMins;

    // Restart within 3 minutes of the target time
    if (deltaMins < 0 || deltaMins > 3)
        return;

    if ( chromiumData.settings.autoRestartMode == "reboot" ) {
        // Reboot device
        m_epad.invokeEPADInfo(
        {
            destination: 'com.exor.EPAD',
            interface: 'com.exor.EPAD',
            path: '/',
            member: 'restartSystemWithImage',
            signature: 'i',
            body: [ 0 ]
        });
    } else {
        // Restart chromium
        m_dbus.sysBus().invoke(
            {
                destination: 'com.exor.JMLauncher',
                interface: 'com.exor.JMLauncher',
                path: '/',
                member: 'restartApp',
                signature: 's',
                body: [ "Chromium" ]
            },
            function (err, eres)
            {
                m_utils.error("Chromium: failed to restart " + err + eres);
            }
        );

        startTime = Date.now();
    }

}

function init (modules, cb, jmu_require)
{
    m_cache = jmu_require('../cache');
    m_rest = jmu_require('../api/rest');
    m_utils = jmu_require('../utils');
    m_epad = jmu_require('../epad');
    m_dbus = jmu_require('../api/dbus');

    m_utils.trace('Initializing Chromium App Module');

    chromiumUrl = m_rest.url + '/' + moduleName;

    client.on('data', function(data) {
        m_utils.trace('Received: ' + data);
        var obj;

        try {
            obj = JSON.parse(data);
        } catch(e) {
            m_utils.error("[Chromium] Received invalid data!");
            return;
        }

        if (obj.data != null) {
           chromiumData = obj.data;
           if (chromiumData.settings.enableAutoRestart && !autoRebootInterval) {
               autoRebootInterval = setInterval(autoRebootHandle, 60*1000);
           } else if (!chromiumData.settings.enableAutoRestart && autoRebootInterval) {
               clearInterval(autoRebootInterval);
               autoRebootInterval = null;
           }
	}

        if (obj.status != null) {
            cmdStatus = obj.status;
            cmdResult = obj.result;
            responseEvent.emit('cmdResult');
        }
    });

    client.on('error', function(err) {
        m_utils.trace("[Chromium] Connection to socket failed");
        client.destroy();
    });

    client.on('close', function() {
        setTimeout(function(){
            client.connect(chromiumPort, '127.0.0.1');
        }, reconnectTimeout);
    });

    client.on('connect', function() {
        m_utils.info('[Chromium] Connected to service socket');
    });

    function sendTCP(data, cb) {

        if (!client || client.pending ) {
            if (cb)
                cb({ error : "Connection not ready" });
            return;
        }

        cmdStatus = status.PENDING;
        if (!client.write(data)) {
            stream.once('drain', cb);
        } else if(cb)
            cb(null);
    }

    var $ = this;

    cacheClear();

    $.get = function (cache, cbt)
    {
        m_utils.trace('get() on /' + moduleName + ' - cache: ' + cache);

	if (chromiumData == null) {
            // Data not available
            if(cbt)
                cbt(null, {});
            return;
        }

        if (!m_cache[$.type])
            m_cache[$.type] = {};

        chromiumData.clockFault = new Date().getYear() < 123;

        m_cache[$.type].chromium = chromiumData;
        if (cbt)
            cbt(null, chromiumData);
    };

    function routeGet (req, res)
    {
        m_rest.routeGet(req, res, $.get);
    };

    function routePost (req, res)
    {
        m_utils.trace('Received POST on /' + moduleName + ': ' + JSON.stringify(req.body));

        if (client.pending || client.destroyed )
            return res.status(404).json({});

        var settings = {};
        if ( req.body.homepage ) {
            if ( req.body.homepage == "" )
               settings.homepage = "";
            else if ( req.body.homepage.indexOf("://") == -1 )
               settings.homepage = "http://" + req.body.homepage;
            else
               settings.homepage = req.body.homepage;
        }

        if ( req.body.onStartup ) {
            settings.onStartup = req.body.onStartup;
            if ( settings.onStartup == "homepage" && settings.homepage == "" )
               settings.onStartup = "settings";
        }

        if ( req.body.enableNavBar != null )
            settings.enableNavBar = req.body.enableNavBar;

        if ( req.body.enableVKeyboard != null )
            settings.enableVKeyboard = req.body.enableVKeyboard;

        if ( req.body.autoNavBar != null )
            settings.autoNavBar = req.body.autoNavBar;

        if ( req.body.historyBtns != null )
            settings.historyBtns = req.body.historyBtns;

        if ( req.body.toolbarBookmarks != null )
            settings.toolbarBookmarks = req.body.toolbarBookmarks;

        if ( req.body.overscrollHistoryNavigation != null )
            settings.overscrollHistoryNavigation = req.body.overscrollHistoryNavigation;

        settings.customUserAgent = req.body.customUserAgent;

        if ( req.body.fallbackPage != null ) {
            if ( req.body.fallbackPage == "" )
               settings.fallbackPage = "";
            else if ( req.body.fallbackPage.indexOf("://") == -1 )
               settings.fallbackPage = "http://" + req.body.fallbackPage;
            else
               settings.fallbackPage = req.body.fallbackPage;
        }

        if ( req.body.alternativeHomePage != null ) {
            if ( req.body.alternativeHomePage == "" )
               settings.alternativeHomePage = "";
            else if ( req.body.alternativeHomePage.indexOf("://") == -1 )
               settings.alternativeHomePage = "http://" + req.body.alternativeHomePage;
            else
               settings.alternativeHomePage = req.body.alternativeHomePage;
        }

        if ( req.body.holdOptionsTimer != null ) {
           if (req.body.holdOptionsTimer > 10)
              req.body.holdOptionsTimer = 10;
           if (req.body.holdOptionsTimer < 3)
              req.body.holdOptionsTimer = 3;
           settings.holdOptionsTimerMilliseconds = req.body.holdOptionsTimer *1000;
        }

        if ( req.body.prefs != null && req.body.prefs.services != null ) {
           settings.prefs = { services: {} };
           if ( req.body.prefs.services.passwordSavingEnabled != null )
              settings.prefs.services.passwordSavingEnabled = req.body.prefs.services.passwordSavingEnabled;
           if ( req.body.prefs.services.downloadRestrictions != null )
              settings.prefs.services.downloadRestrictions = req.body.prefs.services.downloadRestrictions;
           if ( req.body.prefs.services.autofillEnabled != null )
              settings.prefs.services.autofillEnabled = req.body.prefs.services.autofillEnabled;
	}

        if ( req.body.clearCertificateDecisions != null )
          settings.clearCertificateDecisions = req.body.clearCertificateDecisions;

        if ( req.body.enableAutoRestart != null )
          settings.enableAutoRestart = req.body.enableAutoRestart;

	if ( req.body.autoRestartMode == "appRestart" || req.body.autoRestartMode == "reboot" )
          settings.autoRestartMode = req.body.autoRestartMode;

	if ( req.body.autoRestartTimeMins >= 0 && req.body.autoRestartTimeMins < 60*24 )
          settings.autoRestartTimeMins = req.body.autoRestartTimeMins;

	if ( req.body.zoom >= 75 && req.body.zoom <= 150 )
          settings.zoom = req.body.zoom;

        responseEvent.removeAllListeners('cmdResult');
        responseEvent.once('cmdResult', function() {
            if ( cmdStatus != status.OK )
                return res.status(404).json({});
            return res.json(cmdResult)
        });

        sendTCP(JSON.stringify({ cmd: "setSettings", settings: settings}), function(err) {
            if (err) {
                m_utils.error('POST /' + moduleName + ': ' + err.message);
                return res.status(404).json(err);
            }
        });
    };

    function routeCredentialsGet(req, res)
    {
        m_epad.invokeEPADProperty(
            {
                destination: 'com.exor.EPAD',
                path: '/ChromiumSettings',
                interface: 'com.exor.EPAD.ChromiumSettings',
                member: 'getAllCredentialsJson',
            }, function(err, data)
            {
                if (err || data === "")
                {
                    res.status(404).send();
                }
                else
                {
                    json = JSON.parse(data);
                    res.json(json);
                }
            });
    };

    function routeCredentialsPost(req, res)
    {
        m_utils.trace('Received POST on /system/chromium/credentials');

	if (req.body == null)
        {
		res.status(500).send("No user configuration specified");
		return;
	}
        var json = JSON.stringify(req.body);

        m_epad.invokeEPADProperty(
        {
            destination: 'com.exor.EPAD',
            path: '/ChromiumSettings',
            interface: 'com.exor.EPAD.ChromiumSettings',
            member: 'setCredentialsJsonById',
            signature: "is",
            body: [ 0, json ]
        }, function(err, data)
        {
            if (err)
            {
                m_utils.error("Error in setCredentialsJsonById:" + JSON.stringify(err));
                res.status(500).send();
            }
            else if (data < 0)
            {
                m_utils.error("Error code returned setCredentialsJsonById: " + data);
                res.status(400).send("Error code: " + data);
            }
            else
            {
                res.status(200).send();
            }
        });
    };

    function routeCredentialsIdGet(req, res)
    {
        var cid = req.params.cid;
        m_utils.trace('Received GET on /system/chromium/credentials/' + cid + ' (unimplemented)');

        if (isNaN(cid))
        {
		res.status(400).send("Invalid user id" + cid);
		return;
	}

        m_epad.invokeEPADProperty(
        {
            destination: 'com.exor.EPAD',
            path: '/ChromiumSettings',
            interface: 'com.exor.EPAD.ChromiumSettings',
            member: 'getCredentialsJsonById',
            signature: "i",
            body: [ cid ]
        }, function(err, data)
        {
            if (err)
            {
                m_utils.error("Error in getCredentialsJsonById:" + JSON.stringify(err));
                res.status(500).send();
            }
            else if (data === "")
            {
                m_utils.error("Empty reply in getCredentialsJsonById");
                res.status(404).send();
            }
            else
            {
                json = JSON.parse(data);
                res.json(json);
            }
        });
    };

    function routeCredentialsIdPut(req, res)
    {
        var cid = req.params.cid;
        m_utils.trace('Received PUT on /system/chromium/credentials/' + cid + ' (unimplemented)');

        if (isNaN(cid))
        {
		res.status(400).send("Invalid user id" + cid);
		return;
	}

	if (req.body == null)
        {
		res.status(500).send("No user configuration specified");
		return;
	}
        var json = JSON.stringify(req.body);

        m_epad.invokeEPADProperty(
        {
            destination: 'com.exor.EPAD',
            path: '/ChromiumSettings',
            interface: 'com.exor.EPAD.ChromiumSettings',
            member: 'setCredentialsJsonById',
            signature: "is",
            body: [ cid, json ]
        }, function(err, data)
        {
            if (err)
            {
                m_utils.error("Error in setCredentialsJsonById:" + JSON.stringify(err));
                res.status(500).send();
            }
            else if (data < 0)
            {
                m_utils.error("Error code returned setCredentialsJsonById: " + data);
                res.status(400).send("Error code: " + data);
            }
            else
            {
                res.status(200).send();
            }
        });
    };

    function routeCredentialsIdDelete(req, res)
    {
        var cid = req.params.cid;
        m_utils.trace('Received DELETE on /system/chromium/credentials/' + cid + ' (unimplemented)');

        if (isNaN(cid))
        {
		res.status(400).send("Invalid user id" + cid);
		return;
	}

        m_epad.invokeEPADProperty(
        {
            destination: 'com.exor.EPAD',
            path: '/ChromiumSettings',
            interface: 'com.exor.EPAD.ChromiumSettings',
            member: 'deleteCredentialsById',
            signature: "i",
            body: [ cid ]
        }, function(err, data)
        {
            if (err)
            {
                m_utils.error("Error in deleteCredentialsJsonById:" + JSON.stringify(err));
                res.status(500).send();
            }
            else if (data < 0)
            {
                m_utils.error("Error code returned deleteCredentialsJsonById: " + data);
                res.status(404).send("Error code: " + data);
            }
            else
            {
                res.status(200).send();
            }
        });
    };

    function routeCsvCredentialsGet(req, res)
    {
        m_utils.error('Received GET on /system/chromium/csv');

        m_epad.invokeEPADProperty(
            {
                destination: 'com.exor.EPAD',
                path: '/ChromiumSettings',
                interface: 'com.exor.EPAD.ChromiumSettings',
                member: 'exportCredentialsCsv',
            }, function(err, data)
            {
                if (err)
                {
                    m_utils.error("Error in exportCredentialsCsv:" + JSON.stringify(err));
                    res.status(500).send();
                }
                else if (data === "")
                {
                    m_utils.error("Empty reply in exportCredentialsCsv");
                    res.status(500).send();
                }
                else
                {
                    if (req.query.cmd == 'local') {

                        if ( req.query.filename == null )
                            return res.status(500).send();

			m_fs.writeFile(req.query.filename, data, function(err) {
                            if (err)
                                return res.status(500).send();
			    return res.status(200).send();
                        });

                    } else {
                        res.set('Content-Type', 'application/csv')
                            .set('Content-Disposition','attachment;filename=export.csv')
                            .status(200).send(data);
                    }
                }
            });
    };

    function csvImport(res, csvData) {

        var headers = "name,url,username,password".split(',');
        var contentCsvLines = csvData.split("\n");
        var csv = [];

        if ( contentCsvLines.length < 1 )
            return res.status(500).send();

	var fileHeaders = contentCsvLines[0].trim().split(",");
        if ( fileHeaders.length < headers.length )
            return res.status(500).send();

        for(var i = 0; i < headers.length; i++) {
            if ( headers[i] != fileHeaders[i])
                return res.status(500).send();
        }

        for(var i = 1; i < contentCsvLines.length; i++) {
            var obj = {};
            var currentLine = contentCsvLines[i].trim().split(",");
            if (headers.length != currentLine.length) continue;

            for(var j = 0; j < headers.length; j++) {
                obj[headers[j]] = currentLine[j];
            }
            csv.push(obj);
        }

        m_epad.invokeEPADProperty(
        {
            destination: 'com.exor.EPAD',
            path: '/ChromiumSettings',
            interface: 'com.exor.EPAD.ChromiumSettings',
            member: 'importCredentialsJson',
            signature: "s",
            body: [ JSON.stringify(csv) ]
        }, function(err, data)
        {
            if (err)
            {
                m_utils.error("Error in importCredentialsCsv:" + err);
                res.status(500).send();
            }
            else if (data === "")
            {
                m_utils.error("Empty reply in importCredentialsCsv");
                res.status(400).send();
            }
            else
            {
                res.status(200).send(data);
            }
        });
    }

    function routeCsvCredentialsPost(req, res)
    {
        m_utils.trace('Received POST on /system/chromium/csv');

        if (req.body.data != null ) {

            csvImport(res, req.body.data);

        } else if ( req.body.filename != null ) {

            m_utils.error('filename ' + req.body.filename );

            m_fs.readFile(req.body.filename, 'utf8', function (err,data) {
                if (err)
                    return res.status(500).send("No data to import specified");
                 csvImport(res, data);
            });

        } else {
            return res.status(500).send("No data to import specified");
        }
    };

    $.registerHttp = function (app)
    {
        m_utils.info('Registering URL: ' + chromiumUrl);

        app.get(chromiumUrl, routeGet);
        app.post(chromiumUrl, routePost);

        var url = m_rest.url + '/system/chromium';

        var urlCredentials = url + '/credentials';
        m_utils.info('Registering URL: ' + urlCredentials);
        app.get(urlCredentials, routeCredentialsGet);
        app.post(urlCredentials, routeCredentialsPost);

        var urlCredentialsId = url + '/credentials/:cid';
        m_utils.info('Registering URL: ' + urlCredentialsId);
        app.get(urlCredentialsId, routeCredentialsIdGet);
        app.put(urlCredentialsId, routeCredentialsIdPut);
        app.delete(urlCredentialsId, routeCredentialsIdDelete);

        var csvCredentials = url + '/csv';
        app.get(csvCredentials, routeCsvCredentialsGet);
        app.post(csvCredentials, routeCsvCredentialsPost);
    };

    function cacheClear ()
    {
        if ( m_cache[$.type])       
            m_cache[$.type].chromium = null;
    };

    $.registerMonitor = function()
    {
        return { fns : [ $.get ], clear : cacheClear };
    };

    $.type = 'apps';

    client.connect(chromiumPort, '127.0.0.1');

    cb(null);

    return $;
};

exports = module.exports = init;

